#!/usr/bin/env python
'''Script for downloading ERA5 reanalysis data in small amounts.

For large data amount download in batch mode, see batch_download.py.

Author: guangzhi XU (xugzhi1987@gmail.com)
Update time: 2021-04-07 11:33:10.
'''

import os
from era5dl import prepareJobDict, retrieveData

# -----------------Data to retrieve-----------------
DRY = True
VARS = ['u_component_of_wind', 'geopotential']
YEARS = range(1979, 1981)
MONTHS = [1, 2]
DAYS = range(1, 3)
HOURS = [0, 6, 12, 18]
TIME_STEP = 'hourly'    # 'hourly' | 'monthly'
LEVEL_TYPE = 'p'   # 'p', 's'
LEVELS = [1000, 975, 950, 925, 900, 875, 850, 825, 800, 775, 750,
          700, 650, 600, 550, 500, 450, 400, 350, 300, 250, 200, 150, 100, 50]
LEVELS = [1000, 975, 950, 925, 900]
AREA = None  # N, W, S, E
AREA = [10, 0, 0, 10]  # N, W, S, E
FORMAT = 'netcdf'


OUTPUTDIR = './'


if __name__ == '__main__':

    job_dict, data_target = prepareJobDict(
        VARS, YEARS, MONTHS, DAYS, HOURS, TIME_STEP, LEVEL_TYPE, LEVELS,
        area=AREA, out_format=FORMAT)
    fileout_name = 'era5_download.%s' % 'era5.nc' if FORMAT == 'netcdf' else 'era5.grb'
    abpath_out = os.path.join(OUTPUTDIR, fileout_name)
    retrieveData(data_target, job_dict, abpath_out, dry=DRY)



#!/usr/bin/env python
'''Parse web generated API request and split into small jobs.

Author: guangzhi XU (xugzhi1987@gmail.com)
Update time: 2021-04-08 10:10:34.
'''

from __future__ import print_function
from era5dl import batchDownloadFromWebRequest

#--------------Globals------------------------------------------
API_FILE='./examples/test_api1.txt'
OUTPUTDIR = '/home/guangzhi/datasets/era5/newdownload/'
DRY=False




#-------------Main---------------------------------
if __name__=='__main__':

    #batchDownloadFromWebRequest(API_FILE, OUTPUTDIR, ['variable', 'pressure_level', 'year'],
    batchDownloadFromWebRequest(API_FILE, OUTPUTDIR, ['variable', 'year'],
        DRY, pause=3, naming_func=None, verbose=True)



'''Utilities for downloading ECMWF ERA5 reanalysis data.

Author: guangzhi XU (xugzhi1987@gmail.com)
Update time: 2021-04-07 11:29:53.
'''

from __future__ import print_function
import os
import copy
import json
import time
import logging
import logging.config
from pprint import pprint
import cdsapi
from .util_general import get1stOrList, toList, getAttrProduct
from . import util_read_param_table
from . import util_request_parser


# logger config
LOG_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '<%(filename)s-%(funcName)s()>: %(asctime)s,%(levelname)s: %(message)s'},
    },
    'handlers': {
        'default': {
            'level': 'INFO',
            'class': 'logging.FileHandler',
            'formatter': 'standard',
            'filename': 'batch_era5_download.log'
        },
    },
    'loggers': {
        'n1': {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': True
        }
    }
}

__all__=[
        'retrieveData', 'getLogger', 'skipJobs', 'loadDownloadedList',
        'prepareJobDict', 'prepareBatchJobDicts', 'processJob',
        'processJobs', 'batchDownload', 'batchDownloadFromWebRequest',
        'TEMPLATE_DICT'
        ]

TEMPLATE_DICT = {
    'data_target': 'reanalysis-era5-pressure-levels',
    'product_type': 'reanalysis',
    'format': 'netcdf',
    'variable': [
        'geopotential', 'specific_humidity',
    ],
    'pressure_level': [
        '300', '500', '700',
    ],
    'year': [
        '1991', '1992', '1993',
    ],
    'month': [
        '01', '02', '03',
        '04', '05', '06',
        '07', '08', '09',
        '10', '11', '12',
    ],
    'day': [
        '01', '02', '03',
        '04', '05', '06',
        '07', '08', '09',
        '10', '11', '12',
        '13', '14', '15',
        '16', '17', '18',
        '19', '20', '21',
        '22', '23', '24',
        '25', '26', '27',
        '28', '29', '30',
        '31',
    ],
    'time': [
        '00:00', '06:00', '12:00',
        '18:00',
    ],
    'area': [10, 80, -10, 100],
}


def retrieveData(data_target, job_dict, abpath_out, dry=True):
    '''Send cdsapi retrieval request.

    Args:
        data_target (str): target dataset. This is the 1st input arg to the
            cdsapi.Client().retrieve() method.
        job_dict (dict): dictionary describing the data retrieval task.
        abpath_out (str): absolute path to save downloaded data. Create folder if
            not exists already.
    Keyword Args:
        dry (bool): if True, only print the request job without submitting it.

    NOTE that all requirest data are saved into a single file. If data size
    is big, consider break into smaller pieces, using for instance a for-loop.
    '''

    outputdir = os.path.dirname(abpath_out)
    if not os.path.exists(outputdir):
        os.makedirs(outputdir)

    # -------------Run retrieval or dry run-------------
    if dry:
        print('\n########### DRY RUN ############\n')
        print('job_dict = ')
        pprint(job_dict)
        print('data_target = ', data_target)
        print('\nSave file to:', abpath_out)
    else:
        c = cdsapi.Client()
        c.retrieve(data_target, job_dict, abpath_out)

    return


def getLogger(idx, filename, config_base):
    '''Get a new logger for job indexed idx

    Args:
        idx (int): numerical id for the logger.
        filename (str): absolute file path for log file.
        config_base (dict): basic log config to modify.
    Returns:
        logger (logger): logger.
    '''

    logger_name = '%s-%s' % (__name__, str(idx))
    logger_dict = {
        logger_name: {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': True
        }}

    config_base['loggers'] = logger_dict
    config_base['handlers']['default']['filename'] = filename
    logging.config.dictConfig(config_base)

    return logging.getLogger(logger_name)


def skipJobs(job_list, skip_list, down_list):
    '''Remove certain jobs from given job list

    Args:
        job_list (list): list of tuples, each defining some aspects of a data
            retrieval task. E.g.
            (('vars', 'u_component_of_wind'), ('years', 1997), ('pressure_levels', 1000))
        skip_list (list): list of tuples, jobs to skip.
        down_list (list): list of tuples, finished jobs.
    Returns:
        result (list): list of tuples, jobs defined in <job_list>, with those
            skipped (in <skip_list) or finished (<down_list>) removed.
    '''

    skip_list2 = []
    for dii in skip_list:
        lii = getAttrProduct(dii)
        skip_list2.extend(lii)

    # skip jobs from down_list
    job1 = job_list[0]
    fields = [ii[0] for ii in job1]

    skip_list3 = []
    for dii in down_list:
        lii = tuple([(kk, dii[kk]) for kk in fields])
        skip_list3.append(lii)

    skip_list2.extend(skip_list3)
    result = list(set(job_list).difference(skip_list2))

    print('\n# <util_downloader>: Number of jobs defined: %d' % len(job_list))
    print(
        '# <util_downloader>: Number of skipped jobs from <skip_list>: %d' %
        len(skip_list2))
    print(
        '# <util_downloader>: Number of already downloaded jobs: %d' %
        len(skip_list3))
    print('# <util_downloader>: Number of jobs after skipping: %d' % len(result))

    return result


def loadDownloadedList(abpath_in):
    '''Load list of downloaded jobs

    Args:
        abpath_in (str): absolute path to the file containing fininished jobs.
    Returns:
        down_list (list): a list of dicts, each defines a downloaded job. This dict
            is the 2nd input arg to the cdsapi.Client().retrieve() method.
    '''

    if os.path.exists(abpath_in):
        try:
            down_list = [json.loads(l) for l in open(abpath_in, 'r')]
        except ValueError:
            down_list = []
    else:
        down_list = []

    return down_list


def prepareJobDict(var_list, years, months, days, hours, time_step, level_type,
                   levels, area=None, out_format='netcdf'):
    '''Prepare retrieval job dictionary.

    Args:
        var_list (list or tuple or str): name(s) of variable(s) to be retrieved.
        years (list or tuple or int): years. E.g. range(1980, 1990).
        months (list or tuple or int): months. E.g. [12, 1, 2].
        days (list or tuple or int): days. E.g. range(1, 31).
        hours (list or tuple or int): hours. E.g. [0, 6, 12, 18].
        time_step (str): 'hourly': for sub-daily resolution, or 'monthly' for
            monthly averages.
        level_type (str): 'p' for pressure level data, 's' for single level data.
        levels (list or tuple or int): pressure levels. Ignored if <level_type>
            is 's'.
    Keyword Args:
        area (list or tuple): domain to retrieve in the format of [N, W, S, E],
            in degrees of latitutde/longitude. If None (default), retrieve
            global data.
        out_format (str): save 'netcdf' (default) or 'grib' format.
    Returns:
        job_dict (dict): dictionary describing the data retrieval task.
        data_target (str): type of data to retrieve. Currently one of

            'reanalysis-era5-single-levels'
            'reanalysis-era5-pressure-levels'
            'reanalysis-era5-single-levels-monthly-means'
            'reanalysis-era5-pressure-levels-monthly-means'

    This function constructs a dictionary ( < job_dict > ) describing a data
    for the retrieval task. This is dict is used as the 2nd input arg to the
    cdsapi.Client().retrieve() method.

    Not too useful.
    '''

    # ------------Convert to list or string------------
    levels = toList(levels)
    levels.sort()
    levels_str = [str(ii) for ii in levels]
    levels_str = get1stOrList(levels_str)

    years = toList(years)
    years_str = [str(ii) for ii in years]
    years_str = get1stOrList(years_str)

    months = toList(months)
    months = [str(ii).rjust(2, '0') for ii in months]
    months = get1stOrList(months)

    var_list = toList(var_list)

    # ------------------Check variable------------------
    varname_tables = util_read_param_table.readAllTables()

    for vv in var_list:
        if level_type == 's':
            valid_names = [
                dd['Variable name in CDS']
                for dd in util_read_param_table.getSurfaceTables(
                    varname_tables)]
        elif level_type == 'p':
            valid_names = varname_tables['table9']['Variable name in CDS']

        if vv not in valid_names:
            raise Exception(
                "Variable %s is not found in table. Double check the variable name." % vv)

    var_list = get1stOrList(var_list)

    if out_format not in ['netcdf', 'grib']:
        raise Exception("<out_format> can be either 'netcdf' or 'grib'.")

    product_type = 'reanalysis' if time_step == 'hourly' else 'monthly_averaged_reanalysis'

    # ----------------Construct job dict----------------
    job_dict = {
        'product_type': product_type,
        'format': out_format,
        'year': years_str,
        'month': months,
        'variable': var_list,
    }

    if level_type == 'p':
        job_dict['pressure_level'] = levels_str

    if time_step == 'hourly':
        days = toList(days)
        days = [str(ii).rjust(2, '0') for ii in days]
        days = get1stOrList(days)
        job_dict['day'] = days

        hours = toList(hours)
        hours = ['%s:00' % (str(ii).rjust(2, '0')) for ii in hours]
        hours = get1stOrList(hours)
    elif time_step == 'monthly':
        hours = '00:00'

    job_dict['time'] = hours

    if area is not None:
        job_dict['area'] = area

    if time_step == 'hourly':
        if level_type == 's':
            data_target = 'reanalysis-era5-single-levels'
        elif level_type == 'p':
            data_target = 'reanalysis-era5-pressure-levels'
    elif time_step == 'monthly':
        if level_type == 's':
            data_target = 'reanalysis-era5-single-levels-monthly-means'
        elif level_type == 'p':
            data_target = 'reanalysis-era5-pressure-levels-monthly-means'

    return job_dict, data_target


def prepareBatchJobDicts(template_dict, job_dict, skip_list, outputdir,
        naming_func=None):
    '''Prepare a list of job dictionaries for a batch download task.

    Args:
        template_dict (dict): default job dict. Relevant fields will be updated
            using corresponding key-value pairs given in <job_dict>.
        job_dict (dict): dict defining the batch download job. Only need to
            provide attributes that are different from the default values
            given in <template_dict>. Other fields will be taken from
            <template_dict>.

            E.g.
            job_dict = {'variable': ['2m_temperature', '10m_u-component_of_wind'],
                'year': range(1979, 1981),
            }

        skip_list (list): list of dicts, each in the same format as <job_dict>.
            This is used to exclude/skip certain jobs.
        outputdir (str): absolute path to the folder to save downloaded data.
    Keyword Args:
        naming_func (callable or None): if a callable, a function that accepts
            a single input argument which a dict (as <template_dict>) defining
            the data retrieval task, and returns a string as the filename
            (without folder path) to name the downloaded data. If None,
            construct a default filename, using the following format:
                [ID<n>]<attributes>.nc or
                [ID<n>]<attributes>.grb

            where <n> is the numerical id of the job, <attributes> is a
            dash concatenated string joining the attributes that define the job.
            E.g.
                [ID02]700-geopotential-2000.nc
    Returns:
        result (list): a list of dicts, each defines a download job. This dict
            is the 2nd input arg to the cdsapi.Client().retrieve() method.
    '''

    # try get downloaded jobs
    down_list_file = os.path.join(outputdir, 'downloaded_list.txt')
    down_list = loadDownloadedList(down_list_file)

    # form imcomplete job dicts
    jobs = getAttrProduct(job_dict)
    jobs = skipJobs(jobs, skip_list, down_list)
    jobs = [dict(ii) for ii in jobs]

    result = []
    for ii, jobii in enumerate(jobs):

        # get the complete job dict: tmpdictii
        jobid = str(ii).rjust(len(str(len(jobs))), '0')
        tmpdictii = copy.deepcopy(template_dict)
        tmpdictii.update(jobii)

        # ---------------Get output file name---------------
        if naming_func is None:
            keys = list(jobii.keys())
            keys.sort()
            values = [jobii[kk] for kk in keys]
            fileout_name = '[ID%s]%s%s' % (
                jobid, '-'.join(map(str, values)),
                '.nc' if tmpdictii['format'] == 'netcdf' else '.grb')
        else:
            fileout_name = naming_func(tmpdictii)

        abpath_out = os.path.join(outputdir, fileout_name)
        tmpdictii['abpath_out'] = abpath_out

        result.append(tmpdictii)

    return result


def processJob(job_dict, jobid, outputdir, dry):
    '''Process a data retrieval job

    Args:
        job_dict (dict): dict defining a download job. This dict
            is the 2nd input arg to the cdsapi.Client().retrieve() method.
        jobid (str): id of the job.
        outputdir (str): absolute path to the folder to save downloaded data.
        dry (bool): if True, only print the request job without submitting it.

    Function logs the job info to a log file in the specified output directory
    (<outputdir>), and calls the retrieveData() to retrieve data, or simulate
    a dry run.
    '''

    # ---------------Get a logger for job---------------
    logger = getLogger( 'root', os.path.join( outputdir, 'era5_downloader.log'),
        LOG_CONFIG)

    # ---------------------Retrieve---------------------
    data_target = job_dict.pop('data_target')
    abpath_out = job_dict.pop('abpath_out')

    logger.info('<batch_download>: Output folder at: %s' % outputdir)
    logger.info('Launch job %s' % jobid)
    logger.info('Job info: %s' % (str(job_dict)))
    logger.info('Output file location: %s' % abpath_out)

    retrieveData(data_target, job_dict, abpath_out, dry=dry)

    return


def processJobs(job_dicts, outputdir, dry, pause=3, verbose=True):
    '''Process multiple data retrieval jobs

    Args:
        job_dicts (list): a list of dicts, each defines a download job. This dict
            is the 2nd input arg to the cdsapi.Client().retrieve() method.
        outputdir (str): absolute path to the folder to save downloaded data.
        dry (bool): if True, only print the request job without submitting it.
    Keyword Args:
        pause (int): number of seconds to pause between jobs.
    '''

    if len(job_dicts) == 0:
        print('\n# <batch_download>: No job to run.')
    else:
        fail_list = []
        done_list = []
        down_list_file = os.path.join(outputdir, 'downloaded_list.txt')

        for ii, jobii in enumerate(job_dicts):

            idstr = str(ii+1).rjust(len(str(len(job_dicts))), '0')
            print('\n# <batch_download>: Processing job %s/%d\n' %(idstr, len(job_dicts)))

            try:
                processJob(jobii, idstr, outputdir, dry)
            except Exception as e:
                print('Failed job %s.' %idstr, e)
                fail_list.append(jobii)
            else:
                if not dry:
                    done_list.append(jobii)
                    with open(down_list_file, 'a') as down_fout:
                        json.dump(jobii, down_fout)
                        down_fout.write('\n')

                time.sleep(pause)

        # ------------------Print summary------------------
        if len(fail_list) == 0:
            print('\n# <batch_download>: All done.')
        else:
            print('\n# <batch_download>: Failed jobs:')

            for ii in fail_list:
                print(ii)

    return


def batchDownload(template_dict, job_dict, skip_list, outputdir, dry, pause=3,
                  naming_func=None, verbose=True):
    '''Start a batch downloading job

    Args:
        template_dict (dict): default job dict. Relevant fields will be updated
            by corresponding key-value pairs in given in <job_dict>.
        job_dict (dict): dict defining the batch download job. Only need to
            provide attributes that are different from the default template
            given in TEMPLATE_DICT. Other fields will be taken from TEMPLATE_DICT.

            E.g.
            job_dict = {
                'variable': ['2m_temperature', '10m_u-component_of_wind'],
                'year': range(1979, 1981),
            }
        skip_list (list): list of dicts, each in the same format as <job_dict>.
            This is used to exclude/skip certain jobs.
        outputdir (str): absolute path to the folder to save downloaded data.
        dry (bool): if True, only print the request job without submitting it.
    Keyword Args:
        pause (int): number of seconds to pause between jobs.
        naming_func (callable or None): if a callable, a function that accepts
            a single input argument which a dict (as <template_dict>) defining
            the data retrieval task, and returns a string as the filename
            (without folder path) to name the downloaded data. If None,
            construct a default filename, using the following format:
                [ID<n>]<attributes>.nc or
                [ID<n>]<attributes>.grb

            where <n> is the numerical id of the job, <attributes> is a
            dash concatenated string joining the attributes that define the job.
            E.g.
                [ID02]700-geopotential-2000.nc
    '''

    if not os.path.exists(outputdir):
        os.makedirs(outputdir)
        print('\n# <batch_download>: Create folder at: %s' % outputdir)

    jobs = prepareBatchJobDicts(template_dict, job_dict, skip_list, outputdir,
                                naming_func=naming_func)
    processJobs(jobs, outputdir, dry, pause, verbose)

    return


def batchDownloadFromWebRequest(request_file, outputdir, split_fields, dry,
        pause=3, naming_func=None, verbose=True):
    '''Start a batch downloading job split from a web api request

    Args:
        request_file (str): absolute path to a text file containing the data
            retrieval task obtained from the CDS web interface.
        outputdir (str): absolute path to the folder to save downloaded data.
        split_fields (list or tuple): dimensions along which to split the job
            into sub-jobs. E.g. ['variable', 'year', 'pressure_level'] will
            split the retrieval job into a number of sub-jobs such that each
            one retrieves one variable, in one year, on each pressure level.
            If 'variable'=['u', 'v'], 'year'=[1999, 2000], 'pressure_level'=
            [900, 950, 1000], this will create 2x2x3 = 12 sub-jobs, starting
            with ('variable'='u', 'year'=1999, 'pressure_level'=900).
        dry (bool): if True, only print the request job without submitting it.
    Keyword Args:
        pause (int): number of seconds to pause between jobs.
        naming_func (callable or None): if a callable, a function that accepts
            a single input argument which a dict (as <template_dict>) defining
            the data retrieval task, and returns a string as the filename
            (without folder path) to name the downloaded data. If None,
            construct a default filename, using the following format:
                [ID<n>]<attributes>.nc or
                [ID<n>]<attributes>.grb

            where <n> is the numerical id of the job, <attributes> is a
            dash concatenated string joining the attributes that define the job.
            E.g.
                [ID02]700-geopotential-2000.nc
    '''

    if not os.path.exists(outputdir):
        os.makedirs(outputdir)
        print('\n# <batch_download>: Create folder at: %s' % outputdir)

    template_dict = util_request_parser.parseFile(request_file)

    # split jobs
    #jobs=util_request_parser.splitBy(job_dict, split_fields)
    job_dict = dict([(kk, template_dict[kk]) for kk in split_fields])

    jobs = prepareBatchJobDicts(template_dict, job_dict, [], outputdir,
                                naming_func=naming_func)
    processJobs(jobs, outputdir, dry, pause, verbose)

    return



'''Functions for reading ERA5 parameter table info saved in csv format.

Author: guangzhi XU (xugzhi1987@gmail.com)
Update time: 2021-04-09 11:13:20.
'''

from __future__ import print_function
import os
import glob
import csv

TABLE_FOLDER=os.path.abspath(os.path.join(os.path.abspath(__file__), '../tables/'))


def readTable(abpath_in, verbose=True):
    '''Read in a table from a csv file and return table in dict format

    Args:
        abpath_in (str): absolute path to csv file.
    Returns:
        result (dict): table stored in a dict. Keys: table columns. Values:
            a list of column values corresponding to a given header.
    '''

    result={}
    with open(abpath_in, 'r') as fin:
        reader=csv.reader(fin, delimiter=',')
        header=reader.__next__()
        for kii in header:
            result[kii]=[]

        for lii in reader:
            for jj, vjj in enumerate(lii):
                result[header[jj]].append(vjj)

    return result

def queryField(table, target_column, given_column, given_value, verbose=True):
    '''Query from a table a target value in a target column

    Args:
        table (dict): table stored in a dict. Keys: table columns. Values:
            a list of column values corresponding to a given header.
        target_column (str): header name of the target value, defines the column
            of the target value to query.
        given_column (str): header name of the given value, defines the column
            of the given value.
        given_value (str): defines the row of the target value. Note that the
            given_value needs to be able to uniquely define a row.
    Returns:
        result (str): target value, at the row defined by <given_value>, in
            the column specified by <target_column>.
    '''

    if target_column not in table:
        raise Exception("<target_column> '%s' not in <table>.\nAll columns: %s" %(target_column, table.keys()))
    if given_column not in table:
        raise Exception("<given_column> '%s' not in <table>.\nAll columns: %s" %(given_column, table.keys()))

    target_values=table[target_column]
    given_values=table[given_column]
    try:
        idx=given_values.index(given_value)
    except Exception as e:
        raise Exception("<given_value> not found:", e)

    if given_values.count(given_value)>1:
        raise Exception("<given_value> '%s' does not uniquely define a row in <table>." %given_value)

    result=target_values[idx]

    return result

def readAllTables():
    '''Read all tables'''

    table_files=glob.glob(os.path.join(TABLE_FOLDER, 'table*.csv'))
    table_files.sort()
    table_dict={}
    for fii in table_files:
        tableii=readTable(fii)
        nameii=os.path.basename(fii).split('.')[0]
        table_dict[nameii]=tableii

    return table_dict

def getSurfaceTables(all_tables):
    '''Get only tables for surface/single level parameters'''
    return [all_tables['table%d' %ii] for ii in range(1,7)]

def getShortNames(*args):
    '''Get shortName from Variable name in CDS

    Args:
        *args (tuple): tuple of strings, Variable name in CDS.
    Returns:
        results (str or list): corresponding ShortName (e.g. 'z', 't') for each
            variable name in <*args>, ('geopotential', 'temperature').
    '''

    results=[]
    all_tables=readAllTables()

    for vii in args:
        for kjj,tjj in all_tables.items():
            try:
                shortnameii=queryField(tjj, 'shortName', 'Variable name in CDS', vii)
            except:
                shortnameii=vii
            else:
                break
        results.append(shortnameii)

    if len(results)==1:
        return results[0]
    else:
        return results


if __name__=='__main__':

    TABLES=readAllTables()
    a=getShortNames('geopotential', 'divergence', 'relative_humidity', 'runoff', 'aaa')











'''Functions for parse web generated API request and split into small jobs.

Author: guangzhi XU (xugzhi1987@gmail.com)
Update time: 2021-04-08 10:10:34.
'''

from __future__ import print_function
import os
import re
import copy
from pprint import pprint
from .util_general import getAttrProduct

__all__=[
        'DATA_TARGET_PATTERN', 'DICT_PATTERN', 'DICT_KEY_VALUE_PATTERN',
        'parseJobDict', 'splitBy', 'parseFile', 'parseString'
        ]

#------------api request parsing regex------------
DATA_TARGET_PATTERN=re.compile(r'''
^c\.retrieve\(  # start of the retrieve() function
\s*'(.*?)',     # next line
''', re.MULTILINE | re.VERBOSE)

DICT_PATTERN=re.compile(r"^\s*\{(.*)\},", re.DOTALL | re.MULTILINE)

#DICT_KEY_VALUE_PATTERN=re.compile(r"^\s*'(\w+)':\s*(.*?)(?=^\s*'(?:\w+)':|\Z)", re.DOTALL | re.MULTILINE)
DICT_KEY_VALUE_PATTERN=re.compile(r'''
^\s*'(\w+)':            # key, e.g. 'variable':
\s*                     # space
(.*?)                   # value, e.g. 'netcdf', '[\n   '1991', '1992', '1993',\n],
(?=^\s*'(?:\w+)':|\Z)   # non-capturing lookfoward, could be a new key, e.g.
                        # 'year', or the end of the string for the last
                        # key-value pair
''', re.DOTALL | re.MULTILINE | re.VERBOSE)


def parseJobDict(job_list):
    '''Put list of key-value pairs into a dict

    Args:
        job_list (list): list of (key, value) pairs, where <value> is an
            unformatted string, which has to be converted to a normal string,
            or list, using eval(value).
    Returns:
        result (dict): dict containing all key-value pairs in <job_list>.
    '''

    result={}
    for ii, (kii, vii) in enumerate(job_list):
        vii=vii.strip().strip(',')
        valueii=eval(vii)
        result[str(kii)]=valueii

    return result

def splitBy(job_dict, split_fields, verbose=True):
    '''Split total request job into a number of sub-jobs

    Args:
        job_dict (dict): parsed dictionary defining a data retrieval request job.
        split_fields (list or tuple): dimensions along which to split the job
            into sub-jobs. E.g. ['variable', 'year', 'pressure_level'] will
            split the retrieval job into a number of sub-jobs such that each
            one retrieves one variable, in one year, on each pressure level.
            If 'variable'=['u', 'v'], 'year'=[1999, 2000], 'pressure_level'=
            [900, 950, 1000], this will create 2x2x3 = 12 sub-jobs, starting
            with ('variable'='u', 'year'=1999, 'pressure_level'=900).
    Returns:
        results (list): list of dicts, each defines a sub-job retrieval.
    '''

    keys=job_dict.keys()
    split_dims={}
    for kii in split_fields:
        if kii not in keys:
            raise Exception("Key '%s' not in <job_dict>." %kii)

        # don't split area
        if kii in ['area',]:
            if verbose:
                print('\n# <splitBy>: Skip "%s"' %kii)

        split_dims[kii]=job_dict[kii]

    sub_jobs=getAttrProduct(split_dims)
    results=[]
    for jobii in sub_jobs:
        jobii=dict(jobii)
        dictii=copy.deepcopy(job_dict)
        dictii.update(dict(jobii))
        results.append(dictii)

    if verbose:
        print('\n# <splitBy>: Splited job by', split_fields)
        print('# <splitBy>: Number of sub-jobs = %d' %(len(results)))

    return results

def parseFile(abpath_in, verbose=True):
    '''Parse a web generated API request saved in a file

    Args:
        abpath_in (str): absolute path to the text file saving the API request.
    Returns:
        job_dict (dict): a dictionary defining the data retrieval task. This
            is the same form as the 2nd input argument to the
            cdsapi.Client().retrieve() method.
    '''

    with open(abpath_in, 'r') as fin:
        lines=fin.readlines()
        if len(lines)==0:
            raise Exception("File %s is empty." %abpath_in)
        lines=''.join(lines)

    return parseString(lines, verbose)

def parseString(lines, verbose=True):
    '''Parse a web generated API request in a string

    Args:
        lines (str): API request in string format.
    Returns:
        job_dict (dict): a dictionary defining the data retrieval task. This
            is the same form as the 2nd input argument to the
            cdsapi.Client().retrieve() method.
    '''

    try:
        data_target=DATA_TARGET_PATTERN.search(lines).group(1)
    except:
        raise Exception("DATA_TARGET_PATTERN failed to match the string.")

    try:
        job_dict_str=DICT_PATTERN.search(lines).group(1)
    except:
        raise Exception("DICT_PATTERN failed to match the string.")

    try:
        job_dict_pairs=DICT_KEY_VALUE_PATTERN.findall(job_dict_str)
    except:
        raise Exception("DICT_KEY_VALUE_PATTERN failed to match the string.")

    job_dict=parseJobDict(job_dict_pairs)
    job_dict['data_target']=data_target

    return job_dict




#-------------Main---------------------------------
if __name__=='__main__':

    API_FILE='../examples/test_api1.txt'
    OUTPUTDIR='.'

    api_file=os.path.abspath(API_FILE)
    job_dict=parseFile(api_file)
    pprint(job_dict)

    # split jobs
    split_jobs=splitBy(job_dict, OUTPUTDIR, ['variable', 'pressure_level', 'year'])












import setuptools

with open('README.md', 'r') as fin:
    long_description=fin.read()
exec(open('version.py').read())

setuptools.setup(
        name='era5-dl',
        version=__version__,
        author='Guangzhi XU',
        author_email='xugzhi1987@gmail.com',
        description="a simple helper for downloading ECMWF's ERA5 reanalysis data",
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/Xunius/era5-dl',
        packages=setuptools.find_packages(include=['era5dl', ]),
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Intended Audience :: End Users/Desktop',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Natural Language :: English',
            'Operating System :: POSIX :: Linux',
            'Operating System :: MacOS',
            'Operating System :: Microsoft',
            'Programming Language :: Python :: 3',
            'Topic :: Education'
            ],
        install_requires=[
            'cdsapi',
            ],
        python_requires='>=3',
        package_data={'era5dl': ['tables/*.csv', 'examples/*']},
        )




'''General utility functions

Author: guangzhi XU (xugzhi1987@gmail.com)
Update time: 2021-04-17 08:36:48.
'''

from __future__ import print_function
import os
import re

def isListTuple(x):
    """Check an input is a list or tuple or range

    Args:
        x (unknow type): input
    Returns:
        True if <x> is list, tuple or range type, False otherwise.
    """
    import sys

    if sys.version_info.major == 2:
        return isinstance(x, (list, tuple))
    else:
        return isinstance(x, (list, tuple, range))


def get1stOrList(a):
    '''Cast input into a list, or return its 1st element'''
    if len(a) == 1:
        return a[0]
    else:
        return list(a)


def toList(x):
    '''Put input into a list if not already a list'''
    if not isListTuple(x):
        return [x, ]
    else:
        return x


def autoRename(abpath):
    '''Auto rename a file to avoid overwriting an existing file

    <abpath>: str, absolute path to a folder or a file to rename.

    Return <newname>: str, new file path.
    If no conflict found, return <abpath>;
    If conflict with existing file, return renamed file path,
    by appending "_(n)".
    E.g.
        n1='~/codes/tools/send2ever.py'
        n2='~/codes/tools/send2ever_(4).py'
    will be renamed to
        n1='~/codes/tools/send2ever_(1).py'
        n2='~/codes/tools/send2ever_(5).py'
    '''

    def rename_sub(match):
        base = match.group(1)
        delim = match.group(2)
        num = int(match.group(3))
        return '%s%s(%d)' % (base, delim, num+1)

    if not os.path.exists(abpath):
        return abpath

    folder, filename = os.path.split(abpath)
    basename, ext = os.path.splitext(filename)
    # match filename
    rename_re = re.compile('''
            ^(.+?)       # File name
            ([- _])      # delimiter between file name and number
            \\((\\d+)\\) # number in ()
            $''',
                           re.X)

    newname = '%s_(1)%s' % (basename, ext)
    while True:
        newpath = os.path.join(folder, newname)

        if not os.path.exists(newpath):
            break
        else:
            if rename_re.match(newname):
                newname = rename_re.sub(rename_sub, newname)
                newname = '%s%s' % (newname, ext)
            else:
                raise Exception("Exception")

    newname = os.path.join(folder, newname)
    return newname


def getAttrProduct(job_dict):
    '''Create combinations of multiple attributes.

    Args:
        job_dict (dict): dict containing attributes to iterate through.
    Returns:
        result2 (list): list of dicts containing combinations of attributes from
            <job_dict>.
            E.g. job_dict = {'vars' : ['u', 'v'],
                             'years' : [1997, 1998],
                             'levels' : [1000, 900, 800]}
                result2 will be:
                    [{'vars': 'u', 'years': 1997, 'levels': 1000},
                     {'vars': 'u', 'years': 1997, 'levels': 900},
                     {'vars': 'u', 'years': 1997, 'levels': 800},
                     {'vars': 'u', 'years': 1998, 'levels': 1000},
                     {'vars': 'u', 'years': 1998, 'levels': 900},
                     {'vars': 'u', 'years': 1998, 'levels': 800},
                     {'vars': 'v', 'years': 1997, 'levels': 1000},
                     ...
                     {'vars': 'v', 'years': 1998, 'levels': 800}
                     ]
    '''

    pools = [(v,) if not isListTuple(v) else tuple(v)
             for v in job_dict.values()]

    # -----------------Get combinations-----------------
    result = [[]]
    for pool in pools:
        result = [x+[y] for x in result for y in pool]

    result2 = []
    keys = job_dict.keys()
    for ii, jobii in enumerate(result):
        dii = tuple(zip(keys, jobii))
        result2.append(dii)

    return result2




import os
here=os.path.abspath(os.path.dirname(__file__))
abspath=os.path.abspath(os.path.join(here, '../version.py'))
with open(abspath, 'r') as fin:
    exec(fin.read())

from .util_downloader import *
from .util_request_parser import *
